c Fortran 77 algorithms which accompany
c McLeod, A.I. (1997, to appear), Diagnostic check for montone spread,
c Computational Statistics and Data Analysis.
c A.I. McLeod (e-mail: aim@uwo.ca)
c
C dbline
C double-precision version
C calculate regression: y = alpha + beta*x  using biweights
C           and bootstrap distribution of the beta parameter
C
      Subroutine DBLINE(X,Y,N,NBOOT,ISEEDS,RES,DELTA,WS,XSAM,YSAM,ALPHA,
     *    BETA,BOOTECDF,WJACKNIFE,A,IER)
C
C input parameters:
C  x:             vector of length n; log(level)
C  y:             vector of length n; log(spread)
C  n:             number of data values
C  nboot:         number of bootstrap iterations requested
C  res            input vector of dimension n
C  ws             input work vector of dimension n
C  delta          input work vector of dimension n
C  xsam           input work vector of dimension n
C  ysam           input work vector of dimension n
C  wjacknife      input work vector of dimension n
C
C output parameters
C nboot:       number of bootstrap iterations achieved
C res:         residuals from rline regression
C alpha:       intercept parameter
C beta:        slope parameter
C bootecdf:    the bootstrap distribution of alamda (sorted)
C a:           the acceleration, for computing BCa intervals
C ier:         error indicator, ier  = 0; normal
C                                  -1; warning: hs not available
C                                   1; error: n .lt. 9
C                                   2; error: too many ties, xd=0
C                                   3; error: convergence not obtained
C                                   4; error: increase NSTACK:QCKSRTAIM
C
C requires: Wichmann, qcksrt
C
      Implicit Double Precision (A-H,O-Z)
      Real RANDOMAIM, FLOAT
      Dimension X(N), Y(N), RES(N), WS(N), XSAM(N), YSAM(N), ISEEDS(3),
     *    BOOTECDF(NBOOT), DELTA(N), WJACKNIFE(N)
C
       lguout = 9
      IX = ISEEDS(1)
      IY = ISEEDS(2)
      IZ = ISEEDS(3)
C
C fit resistant line regression to logged data
C beta = slope
C alpha = intecept
      Call BLINE(X,Y,N,RES,DELTA,WS,ALPHA,BETA,IERR)
      If (IERR.NE.0) Then
       IER = IERR
       NBOOT = 0
       Return
      Endif
C
C now bootstrap if nboot is >10 otherwise fall thru with defaults
      KBOOT = 0
      If (NBOOT.GT.10) Then
C
C compute the acceleration, a, via jacknifing 
C
       NJACKNIFE = 0
       Do JACKNIFE = 1, N
        JJ = 0
        Do J = 1, N
         If (J.EQ.JACKNIFE) Cycle
         JJ = JJ + 1
         XSAM(JJ) = X(J)
         YSAM(JJ) = Y(J)
        Enddo
        Call BLINE(XSAM,YSAM,(N-1),RES,DELTA,WS,ALPHA_BOOT,BETA_BOOT,
     *      IERR)
        If (IERR.GT.0) Cycle
        NJACKNIFE = NJACKNIFE + 1
        WJACKNIFE(NJACKNIFE) = BETA_BOOT
       Enddo
       AMEAN = 0.0D0
       Do J = 1, NJACKNIFE
        AMEAN = AMEAN + WJACKNIFE(J)
       Enddo
       AMEAN = AMEAN / DFLOAT(NJACKNIFE)
       ATOP = 0.0D0
       ABOT = 0.0D0
       Do J = 1, NJACKNIFE
        ATOP = ATOP + (WJACKNIFE(J)-AMEAN) ** 3
        ABOT = ABOT + (WJACKNIFE(J)-AMEAN) ** 2
       Enddo
       A = ATOP / (6.0D0*DSQRT(ABOT)**3)
c
c now compute the bootstrap distribution
c
       JBOOT = 1 !at this point kboot = 0
       Do While (JBOOT .LE. NBOOT)
        Do J = 1, N
         JJ = 1 + IFIX(FLOAT(N)*RANDOMAIM(IX,IY,IZ))
         XSAM(J) = X(JJ)
         YSAM(J) = Y(JJ)
        Enddo
        Call BLINE(XSAM,YSAM,N,RES,DELTA,WS,ALPHA_BOOT,BETA_BOOT,IERR)
        If (IERR.GT.0) Cycle
        JBOOT = JBOOT+1
        KBOOT = KBOOT + 1
        BOOTECDF(KBOOT) = BETA_BOOT
       Enddo
      Endif
c
c return: number of successful bootstrap iterations
c         seeds
c         bootstrap ecdf
c
      NBOOT = KBOOT
      ISEEDS(1) = IX
      ISEEDS(2) = IY
      ISEEDS(3) = IZ
      Call QCKSRTDAIM(NBOOT,BOOTECDF,IERQ)
      If (IERQ.NE.0) IER=4
      Return
      End
C bisquare straight line fitting
C fast fortran version.
C
      Subroutine BLINE(X,Y,N,RES,DELTA,W,ALPHA,BETA,IERR)
C
C input variables
C  x        x-data, vector of length n
C  y        y-data, vector of length n
C  n        number of data points
C  method   =1, biweights
C           =2, l1-norm
C  res      vector of length n
C  delta    work vector of length n
C  w        work vector of length n
C
C output variables
C  x        vector of length n, explanatory variable
C  y        vector of length n, response
C  res      vector of length n, residuals 
C  delta    vector of length n, final biweights
C  alpha    intercept
C  beta     slope
C  ierr     error indicator, ierr = 0; normal
C                                   1; error: convergence not obtained
c                                   2; error: increase NSTACK:QCKSRTAIM
C
      Implicit Double Precision (A-H,O-Z)
      Parameter (MAXIT = 50)
      Parameter (ERRTOL = 1.0D-5)
      Parameter (ONE = 1.0D0)
      Parameter (TWO = 2.0D0)
      Parameter (SIX = 6.0D0)
      Parameter (ZERO = 0.0D0)
      Dimension X(N), Y(N), RES(N), DELTA(N), W(N)
C
      IERR = 0
      ITER = 1
      ERROR = ONE
C
C median of sorted values of absres(i) is 0.5*(absres(n1)+absres(n2))
      N1 = (N+1) / 2
      If (N.EQ.2*(N/2)) Then
       N2 = N1 + 1
      Else
       N2 = N1
      Endif
C 
      Do I = 1, N
       DELTA(I) = ONE
      Enddo
C
      alpha = zero
      beta  = zero
      Do While ((ITER.LE.MAXIT).AND.(ERROR.GT.ERRTOL))
       ITER = ITER + 1
       XX11 = ZERO
       XX12 = ZERO
       XX22 = ZERO
       XY1 = ZERO
       XY2 = ZERO
       NTOT = 0
       Do I = 1, N
        If (DELTA(I).LE.ZERO) Cycle
        NTOT = NTOT + 1
        XX11 = XX11 + ONE * DELTA(I)
        XX12 = XX12 + X(I) * DELTA(I)
        XX22 = XX22 + X(I) * X(I) * DELTA(I)
        XY1 = XY1 + Y(I) * DELTA(I)
        XY2 = XY2 + X(I) * Y(I) * DELTA(I)
       Enddo
       DET = XX11 * XX22 - XX12 ** 2
       DET1 = XY1 * XX22 - XX12 * XY2
       DET2 = XX11 * XY2 - XY1 * XX12
       alphaprev = alpha
       betaprev  = beta
       ALPHA = DET1 / DET
       BETA = DET2 / DET
       ERROR = DMAX1(DABS(ALPHA-ALPHAPREV),DABS(BETA-BETAPREV))
       Do I = 1, N
        RES(I) = Y(I) - ALPHA - BETA * X(I)
        W(I) = DABS(RES(I))
       Enddo
       Call QCKSRTDAIM(N,W,IERQ)
       If (IERQ.NE.0) Then
        IER = 2
        Return
       Endif
       SCALE = (W(N1)+W(N2)) / TWO
c       ERROR = ZERO
       Do I = 1, N
        DELTAPREV = DELTA(I)
        U = RES(I) / (SIX*SCALE)
        If (DABS(U).LT.ONE) Then
         DELTA(I) = (ONE-U**2) ** 2
        Else
         DELTA(I) = ZERO
        Endif
c        ERROR = DMAX1(ERROR,DABS(DELTA(I)-DELTAPREV))
       Enddo
      Enddo
      Return
      End
      Subroutine WICHMANN(Z,NZ,N,IX,IY,IZ,ITYPE,CHOLCOV,NCHOLCOV,M,WS)
C input parameters
C z  - vector of length nz
C nz - dimension of z, nz >= n*m
C n  - number of replications
C ix,iy,iz  - initial seeds
C itype - =0 uniform(0,1), =1 nid(0,1), =2 multivariate normal
C chol - cholesky decomposition of the covariance matrix stored in symmetric
C        storage form, i.e., if cov = L*(L^transpose), then chol is the
C        nonzero elements of L reading in the following order: first row,
C        second row, third row, etc.
C nchol - dimension of chol (ie. = m*(m+1)/2)
C m  -  m-variate multivariate variables are produced
C ws  -  work vector of dimension m
C
C output parameters
C z  -  vector of random variables. In the case of multivariate normal
C       variables, the first m positions are for the first replication,
C       the next m for the next replication, etc.
C
C references:
C       1. Wichmann, B.A. and Hill, I.D. (1982) "An efficient and 
C          portable random number generator". algorithm as 183.
C          appl. statist. (1982) vol.31, no.2, pp. 188-190.
C       2. McLeod, A.I. (1985) "A remark on algorithm as 183",
C          appl. statist. vol. 34, no. 2, pp. 198--200.
C       3. Box, G.E.P. and Muller, M.E. (1958) "A note on the generation
C          of random normal deviates", annals of mathematical statistics,
C          vol. 29, pp. 610--611.
C
C        returns a pseudo-random number rectangularly distributed
C        between 0 and 1
C
C        ix, iy and iz should be set to integer values between
C        1 and 30000 before first entry
C
C        integer arithmetic up to 30323 is required
C
      Dimension Z(NZ), WS(M), CHOLCOV(NCHOLCOV), C(3), D(3)
      Logical SWTCH
C initialize numerical constants
      Data HALF, ONE, TWOPI, TWONEG /0.5, 1.0, 6.2831853, -2.0/
C coefficients in rational approximation to inverse normal cdf
      Data C /2.515517, 0.802853, 0.010328/
      Data D /1.432788, 0.189269, 0.001308/
      NM = N * M
      Do I = 1, NM
       IX = 171 * MOD(IX,177) - 2 * (IX/177)
       IY = 172 * MOD(IY,176) - 35 * (IY/176)
       IZ = 170 * MOD(IZ,178) - 63 * (IZ/178)
C
       If (IX.LT.0) IX = IX + 30269
       If (IY.LT.0) IY = IY + 30307
       If (IZ.LT.0) IZ = IZ + 30323
C
C        if integer arithmetic up to 5212632 is available,
C        the preceding 6 statements may be replaced by
C
C        IX = MOD(171 * IX, 30269)
C        IY = MOD(172 * IY, 30307)
C        IZ = MOD(170 * IZ, 30323)
C
C        on some machines, this may slightly increase
C        the speed. the results will be identical.
C
       RANDOM = AMOD(FLOAT(IX)/30269.0+FLOAT(IY)/30307.0+FLOAT(IZ)/
     *     30323.0,1.0)
C
       If (RANDOM.EQ.0.0) Then
        RANDOM = DMOD(DBLE(FLOAT(IX))/30269.0D0+DBLE(FLOAT(IY))/
     *      30307.0D0+DBLE(FLOAT(IZ))/30323.0D0,1.0D0)
        If (RANDOM.GE.1.0) RANDOM = 0.999999
       Endif
C
       Z(I) = RANDOM
      Enddo
      If (ITYPE.EQ.0) Return
C
C generate nid(0,1) random variables
      N2 = NM / 2
      II = 1
      Do I = 1, N2
       IIP1 = II + 1
       U = SQRT(TWONEG*ALOG(Z(II)))
       A = TWOPI * Z(IIP1)
       V = SIN(A)
       W = COS(A)
       Z(II) = U * V
       Z(IIP1) = U * W
       II = II + 2
      Enddo
      NN = N2 + N2
C if n odd, use inverse cdf method to obtain the last r.v.
      If (NN.NE.NM) Then
       P = Z(NM)
       SWTCH = .FALSE.
       If (Z(NM).GT.HALF) SWTCH = .TRUE.
       If (SWTCH) P = ONE - Z(NM)
       PSQ = P * P
       T = SQRT(-ALOG(PSQ))
       Z(NM) = T - (C(1)+T*(C(2)+T*C(3))) / (ONE+T*(D(1)+T*(D(2)+T*(
     *     D(3)))))
       If (.NOT.SWTCH) Z(NM) = -Z(NM)
      Endif
      If (ITYPE.EQ.1) Then
       Do I = 1, N
        Z(I) = Z(I) * CHOLCOV(1)
       Enddo
       Return
      Endif
C
C generate multivariate normal
      IJK = 1
      K0 = 0
      Do I = 1, N
       Do K = 1, M
        WS(K) = Z(K0+K)
       Enddo
       K0 = K0 + M
       JK = 1
       Do J = 1, M
        Z(IJK) = 0.0
        Do K = 1, J
         Z(IJK) = Z(IJK) + CHOLCOV(JK) * WS(K)
         JK = JK + 1
        Enddo
        IJK = IJK + 1
       Enddo
      Enddo
      Return
      End
C version: Feb 23 1992
C double precision version
C
      Subroutine QCKSRTDAIM(N,ARR,IER)
      Implicit Double Precision (A-H,O-Z)
      Parameter (M = 7,NSTACK = 50,FM = 7875.0D0,FA = 211.0D0,FC = 
     *    1663.0D0,FMI = 1.2698413D-4)
      Dimension ARR(N), ISTACK(NSTACK)
      IER = 0
      JSTACK = 0
      L = 1
      IR = N
      FX = 0.
      Do While (.TRUE.)
       If (IR-L.LT.M) Then
        Do J = L + 1, IR
         A = ARR(J)
         Do I = J - 1, 1, -1
          If (ARR(I).LE.A) Goto 10
          ARR(I+1) = ARR(I)
         Enddo
         I = 0
   10    ARR(I+1) = A
        Enddo
        If (JSTACK.EQ.0) Return
        IR = ISTACK(JSTACK)
        L = ISTACK(JSTACK-1)
        JSTACK = JSTACK - 2
       Else
        I = L
        J = IR
        FX = MOD(FX*FA+FC,FM)
        IQ = L + (IR-L+1) * (FX*FMI)
        A = ARR(IQ)
        ARR(IQ) = ARR(L)
        Do While (.TRUE.)
         Do While (J.GT.0)
          If (A.GE.ARR(J)) Exit
          J = J - 1
         Enddo
         If (J.LE.I) Goto 20
         ARR(I) = ARR(J)
         I = I + 1
         Do While (I.LE.N)
          If (A.LE.ARR(I)) Exit
          I = I + 1
         Enddo
         If (J.LE.I) Exit
         ARR(J) = ARR(I)
         J = J - 1
        Enddo
        ARR(J) = A
        I = J
        Goto 30
   20   ARR(I) = A
   30   JSTACK = JSTACK + 2
        If (JSTACK.GT.NSTACK) Then
         IER = 1
         RETURN
        Endif
        If (IR-I.GE.I-L) Then
         ISTACK(JSTACK) = IR
         ISTACK(JSTACK-1) = I + 1
         IR = I - 1
        Else
         ISTACK(JSTACK) = I - 1
         ISTACK(JSTACK-1) = L
         L = I + 1
        Endif
       Endif
      Enddo
      End
C version: Feb 23 1992
      Subroutine QCKSRTAIM(N,ARR,IER)
      Parameter (M = 7,NSTACK = 50,FM = 7875.,FA = 211.,FC = 1663.,FMI =
     *    1.2698413E-4)
      Dimension ARR(N), ISTACK(NSTACK)
      IER = 0
      JSTACK = 0
      L = 1
      IR = N
      FX = 0.
      Do While (.TRUE.)
       If (IR-L.LT.M) Then
        Do J = L + 1, IR
         A = ARR(J)
         Do I = J - 1, 1, -1
          If (ARR(I).LE.A) Goto 10
          ARR(I+1) = ARR(I)
         Enddo
         I = 0
   10    ARR(I+1) = A
        Enddo
        If (JSTACK.EQ.0) Return
        IR = ISTACK(JSTACK)
        L = ISTACK(JSTACK-1)
        JSTACK = JSTACK - 2
       Else
        I = L
        J = IR
        FX = MOD(FX*FA+FC,FM)
        IQ = L + (IR-L+1) * (FX*FMI)
        A = ARR(IQ)
        ARR(IQ) = ARR(L)
        Do While (.TRUE.)
         Do While (J.GT.0)
          If (A.GE.ARR(J)) Exit
          J = J - 1
         Enddo
         If (J.LE.I) Goto 20
         ARR(I) = ARR(J)
         I = I + 1
         Do While (I.LE.N)
          If (A.LE.ARR(I)) Exit
          I = I + 1
         Enddo
         If (J.LE.I) Exit
         ARR(J) = ARR(I)
         J = J - 1
        Enddo
        ARR(J) = A
        I = J
        Goto 30
   20   ARR(I) = A
   30   JSTACK = JSTACK + 2
        If (JSTACK.GT.NSTACK) Then
          IER = 1
          RETURN
        Endif
        If (IR-I.GE.I-L) Then
         ISTACK(JSTACK) = IR
         ISTACK(JSTACK-1) = I + 1
         IR = I - 1
        Else
         ISTACK(JSTACK) = I - 1
         ISTACK(JSTACK-1) = L
         L = I + 1
        Endif
       Endif
      Enddo
      Return
      End
C version: Feb 23 1992
      Function RANDOMAIM(IX,IY,IZ)
C
C        ALGORITHM AS 183  APPL. STATIST. (1982) VOL.31, NO.2
C
C        RETURNS A PSEUDO-RANDOMAIM NUMBER RECTANGULARLY DISTRIBUTED
C        BETWEEN 0 AND 1
C
C        IX, IY AND IZ SHOULD BE SET TO INTEGER VALUES BETWEEN
C        1 AND 30000 BEFORE FIRST ENTRY
C
C        INTEGER ARITHMETIC UP TO 30323 IS REQUIRED
C
      IX = 171 * MOD(IX,177) - 2 * (IX/177)
      IY = 172 * MOD(IY,176) - 35 * (IY/176)
      IZ = 170 * MOD(IZ,178) - 63 * (IZ/178)
C
      If (IX.LT.0) IX = IX + 30269
      If (IY.LT.0) IY = IY + 30307
      If (IZ.LT.0) IZ = IZ + 30323
C
C        IF INTEGER ARITHMETIC UP TO 5212632 IS AVAILABLE,
C        THE PRECEDING 6 STATEMENTS MAY BE REPLACED BY
C
C        IX = MOD(171 * IX, 30269)
C        IY = MOD(172 * IY, 30307)
C        IZ = MOD(170 * IZ, 30323)
C
C        ON SOME MACHINES, THIS MAY SLIGHTLY INCREASE
C        THE SPEED. THE RESULTS WILL BE IDENTICAL.
C
      RANDOMAIM = AMOD(FLOAT(IX)/30269.0+FLOAT(IY)/30307.0+FLOAT(IZ)/
     *    30323.0,1.0)
C
      If (RANDOMAIM.GT.0.0) Return
      RANDOMAIM = DMOD(DBLE(FLOAT(IX))/30269.0D0+DBLE(FLOAT(IY))/
     *    30307.0D0+DBLE(FLOAT(IZ))/30323.0D0,1.0D0)
      If (RANDOMAIM.GE.1.0) RANDOMAIM = 0.999999
      Return
      End

